/* Arduino-Direct Arduino Board TEST 
 Tests all Arduino I/O and Analog pins
 terry@arduino-direct.com 
 
 This test assumes that pins 0 (RX Data) and 1 (TX Data) are OK
 if programs can be loaded, and that pin 13 is OK if it's attached LED
 can be blinked. Pin 13 is not tested for current sink ability.
 
 Test progress and pass-fail results are sent to the Serial Monitor.
 At the end of test, Pin 13 LED is blinked very slowly 
 for Pass and quickly for Fail.
 
 This test works with a "Test Shield" that is plugged on top
 of the Arduino under test. This shield has a common isolated 
 connection point. The Common Point has 17 220 ohm resistors 
 connected to every digital pin from 2 to 12, and to every analog
 pin from 0 to 5. It also has two 220,000 ohm resistors which 
 connect it to Ground and +5.0 volts, forming a voltage divider
 that sets the common point to 2.50 volts when all pins are in
 the reset high-impedance input condition. This is read by A0 to
 initially test if any of the pins are "stuck" or "leaking", which 
 will pull the common point away from 2.50 volts
 
 Next, pins 3 thru A5 are tested by setting them High or Low against
 an opposite load presented by pin 2. Each pin must be able to
 source and sink at least 10 ma to bring the common point near 2.50 volts.
 If pin 2 fails, all other pins will fail. 
 
 Next, A0 thru A5 are tested individually with voltages near 0.0, 5.0, 2.5
 and a few others, using the resistors connected to the digital pins to
 make various voltage dividers. 
 */

/*-----( Import needed libraries )-----*/
#include <LED.h>    // Easy HAL blink routines

/*-----( Declare Constants )-----*/

const  int    FirstDigPort =  3;  //Pin 2 is used as reference, but tested at end
const  int    LastDigPort  =  12;
const  int    FirstAnaPort =  15;
const  int    LastAnaPort  =  19;
const  int    A2D_Delay    =  100; //ms delay for A-D to settle

const  float  LowStuckResultOK     =  2.40 ; 
const  float  HighStuckResultOK    =  2.60 ;   
const  float  LowVoltResultOK      =  2.30 ;    
const  float  HighVoltResultOK     =  2.70 ;   

/*-----( Declare objects )-----*/
LED  LED13(13);        // The only LED we can blink!

/*-----( Declare Variables )-----*/

int    DigPort2Test;
int    AnaPort2Test;
int    IntVoltsRead;      // Result of AnalogRead 0..1023
float  FloatVoltsRead;    // Result of AnalogRead 0.0 .. 5.0
int    Failure;          // A failure has occurred

void setup()   /****** SETUP: RUNS ONCE ******/
{
  Serial.begin(9600);
  delay(1000);
  Serial.println("Arduino Board Test V1.06");
  Serial.println("Arduino-Direct.com terry@arduino-direct.com");
  delay(500);
  LED13.blink(300, 4);  // "HI" in Morse code
  delay(800);  
  LED13.blink(300, 2);  

}//--(end setup )---


void loop()   /****** LOOP: RUNS CONSTANTLY ******/
{
  Failure = 0;
  Serial.println();  
  Serial.println("--- TEST START ---");

  Serial.println("TEST 001 - Any stuck pins??");
  delay(500);
  IntVoltsRead = analogRead(A0);
  FloatVoltsRead = a2d2float (IntVoltsRead);
  Serial.print("VoltsRead = ");
  Serial.print(FloatVoltsRead,2);  
  /* 
   Serial.print("LowOK = ");
   Serial.println(LowStuckResultOK,2); 
   Serial.print("High = ");
   Serial.println(HighStuckResultOK,2); 
   */
  if (FloatVoltsRead > LowStuckResultOK
    &&  FloatVoltsRead < HighStuckResultOK)
  {
    Serial.println("  -- OK!");
  }
  else
  {
    Failure ++;
    Serial.println("  ** STUCK OR LEAKY PIN FAIL ** ");    
  }

  /*---------( TEST 2 )----------------------*/
  Serial.println();    
  Serial.println("TEST 002 - Digital Pins Source Current");  
  pinMode(2, OUTPUT);      // Set Load pin to Output mode
  for (  DigPort2Test = FirstDigPort; 
         DigPort2Test <= LastDigPort;
         DigPort2Test ++)
  {//Loop thru Source Tests
    digitalWrite(2, LOW);              // Connect Low Load

    Serial.print("Digital pin ");
//    delay(A2D_Delay);    
    Serial.print(DigPort2Test, DEC);     
    Serial.print(" Source Test ");    
    pinMode(DigPort2Test, OUTPUT);      // Set tested pin to Output mode

    digitalWrite(DigPort2Test, HIGH);  // Pin Sources current HIGH     
    delay(A2D_Delay);
    IntVoltsRead = analogRead(A0);
    FloatVoltsRead = a2d2float (IntVoltsRead);
    pinMode(DigPort2Test, INPUT);      // ReSet tested pin to Input mode   
    Serial.print("VoltsRead = ");
    Serial.print(FloatVoltsRead,2);  

    if (FloatVoltsRead > LowVoltResultOK
      &&  FloatVoltsRead < HighVoltResultOK)
    {
      Serial.println("  -- OK!");
    }
    else
    {
      Failure ++;      
      Serial.println("  ** Source FAIL ** ");    
    }      
  }//END Loop thru Source Tests
  Serial.println();

  /*---------( TEST 3 )----------------------*/

  Serial.println("TEST 003 - Digital Pins Sink Current");  
  pinMode(2, OUTPUT);      // Set Load pin to Output mode

  for (  DigPort2Test = FirstDigPort; 
         DigPort2Test <= LastDigPort;
         DigPort2Test ++)
  {// Loop thru Source Tests
    digitalWrite(2, HIGH);              // Connect HIGH Load

    Serial.print("Digital pin ");   
    Serial.print(DigPort2Test, DEC);     
    Serial.print(" Sink Test ");    
    pinMode(DigPort2Test, OUTPUT);      // Set tested pin to Output mode

    digitalWrite(DigPort2Test, LOW);  // Pin Sources current HIGH     
    delay(A2D_Delay);
    IntVoltsRead = analogRead(A0);
    FloatVoltsRead = a2d2float (IntVoltsRead);
    pinMode(DigPort2Test, INPUT);      // ReSet tested pin to Input mode   
    Serial.print("VoltsRead = ");
    Serial.print(FloatVoltsRead,2);  

    if (FloatVoltsRead > LowVoltResultOK
      &&  FloatVoltsRead < HighVoltResultOK)
    {
      Serial.println("  -- OK!");
    }
    else
    {
      Failure ++;
      Serial.println("  ** Sink FAIL ** ");    
    }      
  }//END Loop thru Sink Tests


  /*---------( TEST 4 )----------------------*/
  Serial.println();    
  Serial.println("TEST 004 - Analog Pins Source Current");  
  pinMode(2, OUTPUT);      // Set Load pin to Output mode
  for (  DigPort2Test = FirstAnaPort; 
         DigPort2Test <= LastAnaPort;
         DigPort2Test ++)
  {//Loop thru Source Tests
    digitalWrite(2, LOW);              // Connect Low Load

    Serial.print("Analog pin ");  
    Serial.print(DigPort2Test - 14, DEC);     
    Serial.print(" Source Test ");    
    pinMode(DigPort2Test, OUTPUT);      // Set tested pin to Output mode

    digitalWrite(DigPort2Test, HIGH);  // Pin Sources current HIGH     
    delay(A2D_Delay);
    IntVoltsRead = analogRead(A0);
    FloatVoltsRead = a2d2float (IntVoltsRead);
    pinMode(DigPort2Test, INPUT);      // ReSet tested pin to Input mode   
    Serial.print("VoltsRead = ");
    Serial.print(FloatVoltsRead,2);  

    if (FloatVoltsRead > LowVoltResultOK
      &&  FloatVoltsRead < HighVoltResultOK)
    {
      Serial.println("  -- OK!");
    }
    else
    {
      Failure ++;      
      Serial.println("  ** Source FAIL ** ");    
    }      
  }//END Loop thru Source Tests
  Serial.println();

  /*---------( TEST 5 )----------------------*/

  Serial.println("TEST 005 - Analog Pins Sink Current");  
  pinMode(2, OUTPUT);      // Set Load pin to Output mode

  for (  DigPort2Test = FirstAnaPort; 
         DigPort2Test <= LastAnaPort;
         DigPort2Test ++)
  {// Loop thru Source Tests
    digitalWrite(2, HIGH);              // Connect HIGH Load

    Serial.print("Analog pin ");  
    Serial.print(DigPort2Test - 14, DEC);     
    Serial.print(" Sink Test ");    
    pinMode(DigPort2Test, OUTPUT);      // Set tested pin to Output mode

    digitalWrite(DigPort2Test, LOW);  // Pin Sources current HIGH     
    delay(A2D_Delay);
    IntVoltsRead = analogRead(A0);
    FloatVoltsRead = a2d2float (IntVoltsRead);
    pinMode(DigPort2Test, INPUT);      // ReSet tested pin to Input mode   
    Serial.print("VoltsRead = ");
    Serial.print(FloatVoltsRead,2);  

    if (FloatVoltsRead > LowVoltResultOK
      &&  FloatVoltsRead < HighVoltResultOK)
    {
      Serial.println("  -- OK!");
    }
    else
    {
      Failure ++;
      Serial.println("  ** Sink FAIL ** ");    
    }      
  }//END Loop thru Sink Tests

  /*---------( TEST 6 )----------------------*/
  Serial.println();
  Serial.println("TEST 006 - Analog Pins A0 to A5: A to D (About 2.50 V)");  
  pinMode(2, OUTPUT);      // Set Load pin to Output mode
  digitalWrite(2, HIGH);              // Connect HIGH Load
  pinMode(3, OUTPUT);      // Set Load pin to Output mode  
  digitalWrite(3, LOW);              // Connect LOW Load  

  for (  AnaPort2Test = FirstAnaPort -1; 
         AnaPort2Test <= LastAnaPort;
         AnaPort2Test ++)
  {// Loop thru Analog Tests
    delay(A2D_Delay);
    IntVoltsRead = analogRead(AnaPort2Test);
    FloatVoltsRead = a2d2float (IntVoltsRead);  
    Serial.print("Analog Test VoltsRead = ");
    Serial.println(FloatVoltsRead,2);  
  }// END Loop thru Analog Tests  


  /*---------( TEST 7 )----------------------*/
  Serial.println();
  Serial.println("TEST 007 - Analog Pins A0 to A5: A to D (About 3.33 V)");  
  pinMode(4, OUTPUT);      // Set Load pin to Output mode
  digitalWrite(4, HIGH);              // Connect 2nd Resistor
  

  for (  AnaPort2Test = FirstAnaPort -1; 
         AnaPort2Test <= LastAnaPort;
         AnaPort2Test ++)
  {// Loop thru Analog Tests
    delay(A2D_Delay);
    IntVoltsRead = analogRead(AnaPort2Test);
    FloatVoltsRead = a2d2float (IntVoltsRead);  
    Serial.print("Analog Test VoltsRead = ");
    Serial.println(FloatVoltsRead,2);  
  }// END Loop thru Analog Tests  


  /*---------( TEST 8 )----------------------*/
  Serial.println();
  Serial.println("TEST 008 - Analog Pins A0 to A5: A to D (About 1.66 V)");  
  pinMode(4, OUTPUT);      // Set Load pin to Output mode
  digitalWrite(4, LOW);              // Connect 2nd Resistor
  

  for (  AnaPort2Test = FirstAnaPort -1; 
         AnaPort2Test <= LastAnaPort;
         AnaPort2Test ++)
  {// Loop thru Analog Tests
    delay(A2D_Delay);
    IntVoltsRead = analogRead(AnaPort2Test);
    FloatVoltsRead = a2d2float (IntVoltsRead);  
    Serial.print("Analog Test VoltsRead = ");
    Serial.println(FloatVoltsRead,2);  
  }// END Loop thru Analog Tests  

  // End Test Digital and Analog Ports: Reset everything
  pinMode(2, INPUT);      // Reset Load pin to Input mode
  pinMode(3, INPUT);      // Reset Load pin to Input mode  
  pinMode(4, INPUT);      // Reset Load pin to Input mode 
  MCUCR |= ( 1 << PUD ); //Disable all Pullups
  
  /*------------( TEST COMPLETE )----------------*/
  Serial.println("TEST COMPLETE");  
  if (Failure > 0)
  {
    Serial.print("!!!!!!!!!!  THERE WERE ");  
    Serial.print(Failure, DEC);            
    Serial.print(" FAILURES !!!!!!!!!!");  
    LED13.blink(150, 50);
  }
  else
  {
    LED13.blink(4000, 4);
  }
  Serial.println();  
//  delay(10000);
  //--(end main loop )---
}
/*-----( Declare User-written Functions )-----*/

/*---( fmap: Converts int Analog readings of 0..1023 to float)---*/
float fmap(float x, float in_min, float in_max, float out_min, float out_max)
{
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
} 

float a2d2float (int raw)
{
  return fmap(raw, 0, 1023, 0.0, 5.0);
}


//*********( THE END )***********







